#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier

from appier_extras import base

class PreflightPart(appier.Part):

    def __init__(self, *args, **kwargs):
        appier.Part.__init__(self, *args, **kwargs)
        self.data = appier.conf("PREFLIGHT_DATA", "")
        self.max_age = appier.conf("PREFLIGHT_MAX_AGE", 86400, cast = int)
        self.data_b = appier.legacy.bytes(self.data, force = True)
        self.max_age_s = str(self.max_age)

    def version(self):
        return base.VERSION

    def load(self):
        appier.Part.load(self)

        self.owner.bind("before_request", self._handler)

    def unload(self):
        appier.Part.unload(self)

        self.owner.bind("before_request", self._handler)

    def _handler(self):
        if not self.owner.request.method == "OPTIONS": return
        if self.owner.request.handled: return
        allow_headers = self.request.get_header("Access-Control-Request-Headers", None)
        if allow_headers: self.request.set_header("Access-Control-Allow-Headers", allow_headers)
        self.request.set_header("Access-Control-Max-Age", self.max_age_s)
        self.owner.request.handle(self.data_b)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier

class Authenticable(appier.Observable):

    @classmethod
    def _unset_account(cls, prefixes = None, safes = [], method = "delete"):
        session = appier.get_session()
        _cls = session.get("cls", None)
        if _cls: cls = appier.get_model(_cls)
        cls._unset_session(prefixes = prefixes, safes = safes, method = method)
        Authenticable.trigger_g(
            "unset_account",
            prefixes = prefixes,
            safes = safes,
            method = method
        )

    @classmethod
    def _unset_session(cls, prefixes = None, safes = [], method = "delete"):
        pass

    def _set_account(self, unset = True, safes = [], method = "set"):
        cls = self.__class__
        self._set_session(unset = unset, safes = safes, method = method)
        Authenticable.trigger_g(
            "set_account",
            self,
            unset = unset,
            safes = safes,
            method = method
        )

    def _set_session(self, unset = True, safes = [], method = "set"):
        pass



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import string

import appier

class SafeFormatter(string.Formatter):

    def __init__(self, fallback = ""):
        self.fallback = fallback

    def get_field(self, field_name, args, kwargs):
        if appier.legacy.PYTHON_3:
            import _string
            first, rest = _string.formatter_field_name_split(field_name)
        else:
            first, rest = field_name._formatter_field_name_split()

        try:
            obj = self.get_value(first, args, kwargs)
        except Exception:
            return self.fallback, first

        for is_attr, key in rest:
            if is_attr:
                try:
                    obj = getattr(obj, key)
                except Exception:
                    obj = self.fallback
                    break
            else:
                try:
                    obj = obj[key]
                except Exception:
                    obj = self.fallback
                    break

        return obj, first



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import time

import appier

from appier_extras import utils

class Prismic(object):
    """
    Partial (mixin) that adds the Prismic CMS proxy functionality
    to a controller or model that inherits from it.

    Most of the methods are meant to be used directly from the
    template or controller.
    """

    _prismic_cache = None
    """ The cache manager that is going to be used store
    prismic values, this should have a huge performance
    impact while accessing the values """

    @classmethod
    def prismic_clear(cls):
        cache = cls._get_prismic_cache()
        cache.clear()

    @classmethod
    def _prismic_deref(cls, entry):
        # verifies if the current value is not a dictionary
        # and that the sys type is not defined in the entry
        # value, if that's the case this is the termination
        # condition and then returns the current entry as the value
        if not isinstance(entry, dict): return entry
        if not "type" in entry: return entry
        if not "value" in entry: return entry

        # retrieves both the type and the (possible) multiple
        # values for the current entry, then retrieves the first
        # value as the "master" value for it
        type = entry["type"]
        values = entry["value"]

        if type == "Text":
            return values

        if type == "Boolean":
            return values

        if type == "Number":
            return values

        if type == "Select":
            return values

        if type == "Image":
            return values["main"]["url"]

        if type == "Link.web":
            return values["url"]

        if type == "StructuredText":
            return "\n".join([value["text"] for value in values])

        if type == "Group":
            group = []
            for value in values:
                value_d = dict([
                    (key, cls._prismic_deref(value))
                    for key, value in appier.legacy.iteritems(value)
                ])
                group.append(value_d)
            return group

        return entry

    @classmethod
    def _get_prismic_cache(cls, serialize = True, ref = None):
        ref = ref or Prismic
        if ref._prismic_cache: return ref._prismic_cache
        cache_engine = appier.conf("CACHE", "memory")
        cache_engine = appier.conf("CMS_CACHE_ENGINE", cache_engine)
        cache_engine = appier.conf("PRISMIC_CACHE_ENGINE", cache_engine)
        cache_engine = cache_engine.capitalize() + "Cache"
        cache_engine = getattr(appier, cache_engine)
        ref._prismic_cache = cache_engine.new(hash = True)
        if serialize: ref._prismic_cache = appier.SerializedCache(cls._prismic_cache)
        return ref._prismic_cache

    @classmethod
    def _filter_params(cls, kwargs):
        params = dict()
        for name in appier.legacy.keys(kwargs):
            if name in ("lang",): continue
            params[name] = kwargs.pop(name)
        return params

    def prismic_objects(self, *args, **kwargs):
        kwargs["multiple"] = True
        return self.prismic_value(*args, **kwargs)

    def prismic_value(
        self,
        key,
        default = None,
        cast = None,
        locale = None,
        timeout = 86400,
        verify = False,
        multiple = False,
        *args,
        **kwargs
    ):
        # retrieves the reference to the parent class value to be used
        # to access class level methods
        cls = self.__class__

        # tries to retrieve the provided locale value and in case it
        # fails uses the locale associated with the current request
        # as a fallback (optimistic approach)
        locale = locale or self.request.locale

        # in case the cast value is set, tries to resolve it into the appropriate
        # cast operation using the default cast methods in the configuration module
        if cast: cast = appier.config._cast_r(cast)

        # in case e the locale value is provided it must be normalized
        # into the format expected by prismic API
        if locale:
            language, territory = locale.split("_", 1)
            locale = language + "-" + territory.upper()
            kwargs["lang"] = locale

        # prefixes the cache's key with locale when retrieving a value for
        # a specific locale, so that proper context is used for caching
        cache_key = locale + ":" + key if locale else key

        # in case there are named arguments provided, assumes that they must
        # be key value pairs for the filtering of the scope and appends the
        # extra filter string to the cache key
        if kwargs: cache_key += ":" + ",".join("%s=%s" % (key, str(value)) for\
            key, value in appier.legacy.iteritems(kwargs))

        # tries to retrieve the reference to the prismic cache engine
        # (singleton instance) and verifies if the key exists in it, returning
        # immediately the value if that's the case
        prismic_cache = cls._get_prismic_cache()
        if cache_key in prismic_cache:
            value = prismic_cache[cache_key]
            if cast and not value == None: value = cast(value)
            return value

        # determines if the provided key references a specific field/value on an
        # object or if it instead tries to retrieve the complete object specification
        # and then selects the appropriate retrieval method
        method = self._prismic_value if "." in key else self._prismic_object
        if multiple: method = self._prismic_objects

        # retrieve the value remotely and sets the value in the cache engine,
        # to avoid further retrievals later on
        value = method(key, default = default, verify = verify, *args, **kwargs)
        prismic_cache.set_item(cache_key, value, expires = time.time() + timeout)

        # runs the casting operation if required and then returns the final
        # value to the caller method
        if cast and not value == None: value = cast(value)
        return value

    def prismic_markdown(
        self,
        key,
        limit = 10,
        default = None,
        verify = False,
        encoding = "utf-8",
        options = dict(
            anchors = False,
            blank = False
        ),
        *args,
        **kwargs
    ):
        value = self.prismic_value(
            key,
            limit = limit,
            default = default,
            verify = verify,
            *args,
            **kwargs
        )
        if not value: return value
        html_b = utils.MarkdownHTML.process_str(
            value,
            options = options
        )
        html_s = html_b.decode(encoding)
        return html_s

    @property
    def prismic_api(self):
        import prismic
        return prismic.API()

    def _prismic_object(self, *args, **kwargs):
        default = kwargs.get("default", None)
        kwargs["limit"] = 1
        objects = self._prismic_objects(*args, **kwargs)
        return objects[0] if objects else default

    def _prismic_objects(
        self,
        key,
        limit = 10,
        default = [],
        verify = False,
        *args,
        **kwargs
    ):
        # retrieves the reference to the parent class value to be used
        # to access class level methods
        cls = self.__class__

        # runs a series of assertions for the provided key, raising exceptions
        # in case one of the pre-conditions is not met
        appier.verify(
            not "." in key,
            message = "Malformed key '%s', must include both document type and key" % key,
            code = 400
        )

        # filters the keyword arguments based arguments retrieving only the valid
        # parameters to be used in the filters extension process
        params = cls._filter_params(kwargs)

        # sets the document type value as the provided key as we're trying
        # to retrieve an object instead of a field
        document_type = key

        # retrieves the complete set of items that meet the document type
        # criteria and selects the first one, default to an empty dictionary
        # in case no items exist for such content type
        try:
            query = ["[[at(document.type,\"%s\")]]" % document_type]
            query.extend(["[[at(my.%s.%s,\"%s\")]]" % (document_type, key, value) for\
                key, value in appier.legacy.iteritems(params)])
            entries = self.prismic_api.search_documents(
                q = query,
                page_size = limit,
                *args,
                **kwargs
            ) or []
        except Exception as exception:
            self.logger.warning("Problem while accessing prismic: %s" % exception)
            return default

        # creates the list that is going to hold the multiple entry
        # maps/objects that are going to store the dereferenced
        # prismic entries (after the filtering)
        entries_m = []

        # iterates over the complete set of entries, to be able to
        # creates the map/object with the dereferenced values
        for entry in entries:
            # creates the entry object/map that is going to be
            # populated with the complete set of dereferenced fields
            entry_m = dict()

            # retrieves the data part of the entry and the
            # field itself according to the document type
            data = entry.get("data", {})
            field = data.get(document_type, {})

            # iterates over the complete set of fields identifiers
            # and values and dereferences the values setting them
            # then in the current entry map/object
            for field_id, field_value in appier.legacy.iteritems(field):
                field_value = cls._prismic_deref(field_value)
                entry_m[field_id] = field_value

            # adds the entry map/object to the current list of entries,
            # this list should represent a simplified structure
            entries_m.append(entry_m)

        # returns the final processed list of entry objects with
        # the complete set of dereferenced values
        return entries_m

    def _prismic_value(
        self,
        key,
        limit = 1,
        default = None,
        verify = False,
        *args,
        **kwargs
    ):
        # retrieves the reference to the parent class value to be used
        # to access class level methods
        cls = self.__class__

        # runs a series of assertions for the provided key, raising exceptions
        # in case one of the pre-conditions is not met
        appier.verify(
            "." in key,
            message = "Malformed key '%s', must include both document type and key" % key,
            code = 400
        )

        # filters the keyword arguments based arguments retrieving only the valid
        # parameters to be used in the filters extension process
        params = cls._filter_params(kwargs)

        # splits the provided key around the dot value (namespace oriented)
        # to obtain the document type and the field id
        document_type, field_id = key.split(".", 1)

        # retrieves the complete set of items that meet the document type
        # criteria and selects the first one, default to an empty dictionary
        # in case no items exist for such content type
        try:
            query = ["[[at(document.type,\"%s\")]]" % document_type]
            query.extend(["[[at(my.%s.%s,\"%s\")]]" % (document_type, key, value) for\
                key, value in appier.legacy.iteritems(params)])
            entries = self.prismic_api.search_documents(
                q = query,
                page_size = limit,
                *args,
                **kwargs
            ) or []
        except Exception as exception:
            self.logger.warning("Problem while accessing prismic: %s" % exception)
            return default

        # retrieves the complete set of items and in case there's at least
        # one returns the first one of it otherwise returns an empty dictionary
        entry = entries[0] if entries else dict()

        # retrieves the complete set of field for the entry and tries to retrieve
        # the requested field (by its identifier)
        data = entry.get("data", {})
        field = data.get(document_type, {})

        # verifies if the requested field exists raising an exception otherwise
        # this should ensure that the value exists in prismic
        if verify:
            appier.verify(
                field_id in field,
                message = "'%s' not found" % field_id,
                exception = appier.NotFoundError
            )

        # retrieves the value of the field requested with the provided identifier
        # defaulting to the provided default value in case it does not exists
        field_value = field.get(field_id, default)

        # runs the dereferencing process for the field value that is going to
        # be retrieved so the proper (expected) value may be returned to caller
        field_value = cls._prismic_deref(field_value)

        # returns the final dereferenced value to the caller method
        # this can be both a plain value or a sequence
        return field_value



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import functools

import appier

def recaptcha_protect(action = "default", name = "recaptcha_token"):

    def decorator(function):

        @functools.wraps(function)
        def interceptor(self, *args, **kwargs):
            token = self.field(name, None)
            recaptcha_ensure(self, token, action = action)
            return appier.call_safe(function, self, *args, **kwargs)
        return interceptor

    return decorator

def recaptcha_ensure(self, token, action = "default", force = False):
    if not _recaptcha_available() and not force: return
    secret = appier.conf("RECAPTCHA_SECRET",  None)
    min_score = appier.conf("RECAPTCHA_MIN", 0.5)
    result = appier.post(
        "https://www.google.com/recaptcha/api/siteverify",
        params = dict(
            secret = secret,
            response = token
        )
    )
    if result.get("score", 0) >= min_score and\
        result.get("action", None) == action: return token
    raise appier.AppierException(
        message = "Invalid reCAPTCHA score or action",
        code = 403
    )

def _recaptcha_available():
    key = appier.conf("RECAPTCHA_KEY",  None)
    secret = appier.conf("RECAPTCHA_SECRET",  None)
    return True if key and secret else False



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import logging

import appier

from appier_extras import base

from . import handler

class LogglyPart(appier.Part):
    """
    Modular part class that provides the system to handle Loggly
    logging requests automatically on Appier.

    This part uses the Loggly API client to run all the remote
    calls accordingly.

    :see: http://loggly.com
    """

    def __init__(self, *args, **kwargs):
        appier.Part.__init__(self, *args, **kwargs)
        self.log = kwargs.get("log", False)
        self.buffer_size = kwargs.get("buffer_size", 128)
        self.timeout = kwargs.get("timeout", 30)
        self.log = appier.conf("LOGGLY_LOG", self.log, cast = bool)
        self.buffer_size = appier.conf(
            "LOGGLY_BUFFER_SIZE",
            self.buffer_size,
            cast = int
        )
        self.timeout = appier.conf(
            "LOGGLY_TIMEOUT",
            self.timeout,
            cast = int
        )

    def version(self):
        return base.VERSION

    def load(self):
        appier.Part.load(self)

        self._api = None
        self.add_handler()

    def add_handler(self, set_default = True):
        if not self.log: return
        api = self._get_api()
        handler_loggly = handler.LogglyHandler(
            owner = self,
            api = api,
            buffer_size = self.buffer_size,
            timeout = self.timeout
        )
        handler_loggly.setLevel(self.owner.level)
        handler_loggly.setFormatter(self.owner.formatter)
        self.owner.handlers.append(handler_loggly)
        self.logger.addHandler(handler_loggly)
        if not set_default: return
        logger = logging.getLogger()
        logger.addHandler(handler_loggly)

    def _get_api(self):
        if self._api: return self._api
        try: loggly = appier.import_pip("loggly", package = "loggly_api_python")
        except Exception: loggly = None
        if not loggly: return None
        self._api = loggly.API()
        return self._api



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier

from appier_extras.parts.admin.models import base

class Locale(base.Base):

    locale = appier.field(
        index = "all",
        default = True
    )

    context = appier.field(
        index = "all"
    )

    data_j = appier.field(
        type = dict,
        private = True,
        meta = "longmap",
        description = "JSON Data"
    )

    count_l = appier.field(
        type = int,
        initial = 0,
        description = "Count"
    )

    @classmethod
    def setup(cls):
        super(Locale, cls).setup()
        cls._flush()
        appier.get_bus().bind("locale/reload", cls._flush)

    @classmethod
    def validate(cls):
        return super(Locale, cls).validate() + [
            appier.not_null("locale"),
            appier.not_empty("locale"),
            appier.is_regex("locale", "^[a-z]{2}(?:_[a-z]{2}(?:_[a-z]+)?)?$"),

            appier.not_null("data_j"),
            appier.not_empty("data_j")
        ]

    @classmethod
    def list_names(cls):
        return ["locale", "context", "description", "count_l"]

    @classmethod
    def bundles_d(cls, locale = None):
        locales = cls.find_e(rules = False)
        return dict([((locale.locale, locale.context or None), locale.data_u) for locale in locales])

    @classmethod
    @appier.operation(
        name = "Import Bundle",
        parameters = (
            ("JSON File", "file", "file"),
            ("Locale", "locale", str),
            ("Context", "context", str)
        ),
        factory = True
    )
    def import_bundle_s(cls, file, locale, context = None, strict = False):
        context = context or None
        data_j = cls._json_read(file)
        conditions = cls._conditions(locale, context = context)
        locale_e = cls.get(rules = False, raise_e = False, **{"$and" : conditions})
        if locale_e: locale_e.data_j.update(data_j)
        else: locale_e = cls(locale = locale, context = context, data_j = data_j)
        locale_e.save()
        return locale_e

    @classmethod
    @appier.operation(
        name = "Import CSV",
        parameters = (("CSV File", "file", "file"),)
    )
    def import_csv_s(cls, file):
        csv_reader = cls._csv_read(file)
        header = next(csv_reader)

        result = dict()
        locales_n = header[1:]
        for line in csv_reader:
            name, locales_v = line[0], line[1:]
            for locale_n, locale_v in zip(locales_n, locales_v):
                if not locale_v: continue
                locale_m = result.get(locale_n, {})
                locale_m[name] = locale_v
                result[locale_n] = locale_m

        for locale, data_j in appier.legacy.iteritems(result):
            conditions = cls._conditions(locale)
            locale_e = cls.get(rules = False, raise_e = False, **{"$and" : conditions})
            if locale_e: locale_e.data_j.update(data_j)
            else: locale_e = cls(locale = locale, data_j = data_j)
            locale_e.save()

    @classmethod
    @appier.link(name = "Export CSV", context = True)
    def list_csv_url(cls, view = None, context = None, absolute = False):
        return appier.get_app().url_for(
            "admin.list_locales_csv",
            view = view,
            context = context,
            absolute = absolute
        )

    @classmethod
    @appier.view(
        name = "Explorer",
        parameters = (("Context", "context", str),)
    )
    def explorer_v(cls, context, *args, **kwargs):
        kwargs["sort"] = kwargs.get("sort", [("id", -1)])
        if context: kwargs.update(context = context)
        return appier.lazy_dict(
            model = cls,
            kwargs = kwargs,
            entities = appier.lazy(lambda: cls.find(*args, **kwargs)),
            page = appier.lazy(lambda: cls.paginate(*args, **kwargs))
        )

    @classmethod
    def _flush(cls):
        # retrieves the dictionary of bundles currently
        # enabled in the data source
        bundles_d = cls.bundles_d()

        # iterates over the complete set of locale data pairs
        # in the bundles dictionary to set these bundles (locales)
        for (locale, context), data_j in appier.legacy.iteritems(bundles_d):
            appier.get_app()._register_bundle(data_j, locale, context = context)

    @classmethod
    def _escape(cls, data_j, target = ".", sequence = "::"):
        data_j = dict(data_j)
        for key, value in appier.legacy.items(data_j):
            _key = key.replace(target, sequence)
            if _key == key: continue
            data_j[_key] = value
            del data_j[key]
        return data_j

    @classmethod
    def _unescape(cls, data_j, target = ".", sequence = "::"):
        return cls._escape(data_j, target = sequence, sequence = target)

    @classmethod
    def _conditions(cls, locale, context = None):
        conditions = []
        conditions.append(dict(locale = locale))
        if context: conditions.append(dict(context = context))
        else: conditions.append({"$or" : [dict(context = ""), dict(context = None)]})
        return conditions

    def pre_create(self):
        base.Base.pre_create(self)
        self.count_l = len(self.data_j)
        self.data_j = self.__class__._escape(self.data_j)

    def pre_update(self):
        appier.Model.pre_update(self)
        self.count_l = len(self.data_j)
        self.data_j = self.__class__._escape(self.data_j)

    def post_create(self):
        base.Base.post_create(self)
        self.owner.trigger_bus("locale/reload")

    def post_update(self):
        base.Base.post_update(self)
        self.owner.trigger_bus("locale/reload")

    def post_delete(self):
        base.Base.post_delete(self)
        self.owner.trigger_bus("locale/reload")

    @appier.operation(
        name = "Set Context",
        parameters = (("Context", "context", str),)
    )
    def set_context_s(self, context):
        locale = self.reload(rules = False)
        locale.context = context
        locale.save()

    @appier.link(name = "Export Bundle")
    def bundle_url(self, absolute = False):
        return appier.get_app().url_for(
            "admin.bundle_locale_json",
            id = self.id,
            absolute = absolute
        )

    @property
    def data_u(self):
        cls = self.__class__
        return cls._unescape(self.data_j)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import unittest

import appier
import appier_extras

class AccountTest(unittest.TestCase):

    def setUp(self):
        self.app = appier.App(
            parts = (appier_extras.admin.AdminPart,),
            session_c = appier.MemorySession
        )

    def tearDown(self):
        self.app.unload()
        adapter = appier.get_adapter()
        adapter.drop_db()

    def test_basic(self):
        account = appier_extras.admin.Account()
        account.username = "username"
        account.email = "username@domain.com"
        account.password = "password"
        account.password_confirm = "password"
        account.save()

        self.assertNotEqual(account.id, None)
        self.assertNotEqual(account.password, "password")
        self.assertEqual(account.username, "username")
        self.assertEqual(account.email, "username@domain.com")
        self.assertEqual(len(account.password), 84)

        account = appier_extras.admin.Account.login("username", "password")

        self.assertNotEqual(account, None)
        self.assertEqual(account.username, "username")
        self.assertEqual(account.email, "username@domain.com")

        account.password = "passwordchanged"
        account.password_confirm = "password"

        self.assertRaises(appier.ValidationError, account.save)

        account.password_confirm = "passwordchanged"
        account.save()

        self.assertRaises(
            appier.OperationalError,
            lambda: appier_extras.admin.Account.login("username", "password")
        )

        account = appier_extras.admin.Account.login("username", "passwordchanged")

        self.assertNotEqual(account, None)
        self.assertEqual(account.username, "username")
        self.assertEqual(account.email, "username@domain.com")

        account.save()

        account = appier_extras.admin.Account.login("username", "passwordchanged")

        self.assertNotEqual(account, None)
        self.assertEqual(account.username, "username")
        self.assertEqual(account.email, "username@domain.com")

        account.password = "P"
        account.password_confirm = "P"

        self.assertRaises(appier.ValidationError, account.save)

    def test_insensitive(self):
        account = appier_extras.admin.Account()
        account.username = "Username"
        account.email = "USERNAME@domain.com"
        account.password = "password"
        account.password_confirm = "password"
        account.save()

        self.assertNotEqual(account.id, None)
        self.assertNotEqual(account.password, "password")
        self.assertEqual(account.username, "username")
        self.assertEqual(account.email, "username@domain.com")
        self.assertEqual(len(account.password), 84)

        account = appier_extras.admin.Account.login("USERNAME", "password")

        self.assertNotEqual(account, None)
        self.assertEqual(account.username, "username")
        self.assertEqual(account.email, "username@domain.com")

        self.assertRaises(
            appier.OperationalError,
            lambda: appier_extras.admin.Account.login("username", "PASSWORD")
        )

    def test_confirm(self):
        account = appier_extras.admin.Account()
        account.username = "username"
        account.email = "username@domain.com"
        account.password = "password"
        account.password_confirm = "password"
        account.save()

        self.assertEqual(account.enabled, True)
        self.assertNotEqual(account.confirmation_token, None)

        account = account.get(username = "username", rules = False)
        account.enabled = False
        account.save()

        appier_extras.admin.Account.confirm(account.confirmation_token)
        account = account.reload(rules = False)

        self.assertEqual(account.enabled, True)
        self.assertEqual(account.confirmation_token, None)
        self.assertEqual(account.password, account.encrypt("password"))

    def test_reset(self):
        account = appier_extras.admin.Account()
        account.username = "username"
        account.email = "username@domain.com"
        account.password = "password"
        account.password_confirm = "password"
        account.save()

        self.assertEqual(account.enabled, True)
        self.assertEqual(account.reset_token, None)
        self.assertNotEqual(account.confirmation_token, None)

        account = account.get(username = "username", rules = False)
        account.enabled = False
        account.save()

        self.assertEqual(account.enabled, False)
        self.assertEqual(account.reset_token, None)
        self.assertNotEqual(account.confirmation_token, None)

        reset_token = account.recover_s()
        account = account.reload(rules = False)

        self.assertEqual(account.enabled, False)
        self.assertNotEqual(account.reset_token, None)
        self.assertEqual(account.reset_token, reset_token)
        self.assertNotEqual(account.confirmation_token, None)

        appier_extras.admin.Account.reset(
            reset_token,
            "passwordnew",
            "passwordnew"
        )
        account = account.reload(rules = False)

        self.assertEqual(account.enabled, True)
        self.assertEqual(account.reset_token, None)
        self.assertEqual(account.confirmation_token, None)
        self.assertEqual(account.password, account.encrypt("passwordnew"))

        account.enabled = False
        account.save()

        reset_token = account.recover_s()
        account = account.reload(rules = False)
        appier_extras.admin.Account.reset(
            reset_token,
            "passwordnew",
            "passwordnew",
            confirm = False
        )
        account = account.reload(rules = False)

        self.assertEqual(account.enabled, False)
        self.assertEqual(account.reset_token, None)
        self.assertEqual(account.confirmation_token, None)
        self.assertEqual(account.password, account.encrypt("passwordnew"))

    def test_role(self):
        account = appier_extras.admin.Account()
        account.username = "username"
        account.email = "username@domain.com"
        account.password = "password"
        account.password_confirm = "password"
        account.save()

        tokens = account.tokens()

        self.assertEqual(account.type, appier_extras.admin.Account.USER_TYPE)
        self.assertEqual(tokens, ["base", "user"])

        role = appier_extras.admin.Role()
        role.name = "admin"
        role.tokens = ["*"]
        role.save()

        self.assertEqual(role.name, "admin")
        self.assertEqual(role.tokens, ["*"])

        account = account.reload()
        account.add_role_s("admin")
        tokens = account.tokens()

        self.assertEqual(account.type, appier_extras.admin.Account.ROLE_TYPE)
        self.assertEqual(len(account.roles), 1)
        self.assertEqual(tokens, ["*"])

        account.add_role_s("admin")
        tokens = account.tokens()

        self.assertEqual(account.type, appier_extras.admin.Account.ROLE_TYPE)
        self.assertEqual(len(account.roles), 1)
        self.assertEqual(tokens, ["*"])

        role = appier_extras.admin.Role()
        role.name = "user"
        role.tokens = ["base", "user"]
        role.save()

        self.assertEqual(role.name, "user")
        self.assertEqual(role.tokens, ["base", "user"])

        account = account.reload()
        account.add_role_s("user")
        tokens = account.tokens()

        self.assertEqual(account.type, appier_extras.admin.Account.ROLE_TYPE)
        self.assertEqual(len(account.roles), 2)
        self.assertEqual(tokens, ["*"])

        account.remove_role_s("admin")
        tokens = account.tokens()

        self.assertEqual(account.type, appier_extras.admin.Account.ROLE_TYPE)
        self.assertEqual(len(account.roles), 1)
        self.assertEqual(tokens, ["base", "user"])

        role = appier_extras.admin.Role()
        role.name = "advanced"
        role.tokens = ["base", "advanced"]
        role.save()

        account.add_role_s("advanced")
        tokens = account.tokens()

        self.assertEqual(account.type, appier_extras.admin.Account.ROLE_TYPE)
        self.assertEqual(len(account.roles), 2)
        self.assertEqual(tokens, ["advanced", "base", "user"])

    def test_duplicate_keys(self):
        account = appier_extras.admin.Account()
        account.username = "Username"
        account.email = "USERNAME@domain.com"
        account.password = "password"
        account.password_confirm = "password"
        account.save()

        account_duplicate = appier_extras.admin.Account()
        account_duplicate.username = "Username2"
        account_duplicate.email = "USERNAME2@domain.com"
        account_duplicate.password = "password2"
        account_duplicate.password_confirm = "password2"
        account_duplicate.save()

        account_duplicate.password = "password2"
        account_duplicate.password_confirm = "password2"
        account_duplicate.key = account.key

        self.assertRaises(
            appier.ValidationError,
            lambda: account_duplicate.save()
        )

    def test__join_m(self):
        account = appier_extras.admin.Account()

        result = account._join_m(
            {
                "name" : "brands-origin",
                "brand" : "nike",
                "ctx" : {
                    "brand" : "nike"
                }
            }, {
                "name" : "brands-target",
                "brand" : "adidas",
                "ctx" : {
                    "brand" : "adidas"
                }
            }
        )

        self.assertEqual(
            result,
            {
                "name" : ["brands-target", "brands-origin"],
                "brand" : ["adidas", "nike"],
                "ctx" : {
                    "brand" : ["adidas", "nike"]
                }
            }
        )

        result = account._join_m(
            {
                "name" : "brands-origin",
                "brand" : "nike",
                "ctx" : {
                    "brand" : ["nike", "nifty"]
                }
            }, {
                "name" : "brands-target",
                "brand" : "adidas",
                "ctx" : {
                    "brand" : ["adidas", "amidas"]
                }
            }
        )

        self.assertEqual(
            result,
            {
                "name" : ["brands-target", "brands-origin"],
                "brand" : ["adidas", "nike"],
                "ctx" : {
                    "brand" : ["adidas", "amidas", "nike", "nifty"]
                }
            }
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier

from appier_extras import base

from . import util

class ReCaptchaPart(appier.Part):
    """
    Modular part class that provides the required infra-structure
    for the control of Google's reCAPTCHA service.

    Should be used with proper knowledge of the inner workings of
    the captcha mechanism to avoid any security problems.

    :see: https://developers.google.com/recaptcha
    """

    def version(self):
        return base.VERSION

    def load(self):
        appier.Part.load(self)

        self.owner.context["recaptcha"] = self.recaptcha
        self.owner.context["recaptcha_input"] = self.recaptcha_input
        self.owner.context["recaptcha_script"] = self.recaptcha_script
        self.owner.context["recaptcha_available"] = self.recaptcha_available

    def recaptcha(self, action = "default", name = "recaptcha_token", force = False):
        if not util._recaptcha_available() and not force: return ""
        return self.owner.escape_template(
            self._build_input(name) + self._build_script(action)
        )

    def recaptcha_input(self, name = "recaptcha_token", force = False):
        if not util._recaptcha_available() and not force: return ""
        return self.owner.escape_template(
            self._build_input(name)
        )

    def recaptcha_script(self, action = "default", force = False):
        if not util._recaptcha_available() and not force: return ""
        return self.owner.escape_template(
            self._build_script(action)
        )

    def recaptcha_available(self):
        return util._recaptcha_available()

    def _build_input(self, name):
        return "<input type=\"hidden\" id=\"recaptcha-token\" name=\"%s\" />" % name

    def _build_script(self, action):
        recaptcha_key = appier.conf("RECAPTCHA_KEY", None)
        appier.verify(recaptcha_key, message = "No reCAPTCHA site key provided")
        return "<script type=\"application/javascript\" src=\"https://www.google.com/recaptcha/api.js?render=%s\"></script>" % recaptcha_key +\
            "<script type=\"application/javascript\">window.genRecaptcha = function() { grecaptcha.ready(function() {" +\
            "document.getElementById(\"recaptcha-token\") && grecaptcha.execute(\"%s\", {action: \"%s\"}).then(function(token) {" % (recaptcha_key, action) +\
            "document.getElementById(\"recaptcha-token\").value = token;" +\
            "});" +\
            "});" +\
            "};" +\
            "window.genRecaptcha();" +\
            "</script>"



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier

from appier_extras.parts.admin.models import base

class Role(base.Base):

    name = appier.field(
        index = "all",
        default = True
    )
    """ The common name to be used to identify this role,
    this is going to be the primary way of identifying it """

    tokens = appier.field(
        type = list
    )
    """ The set of ACL token that are going to be used to
    control the permission of accounts that use this role """

    view_ = appier.field(
        type = dict,
        description = "View"
    )
    """ The filtered view that is going to be applied for
    every filtered operation (data source access) """

    children = appier.field(
        type = appier.references(
            "Role",
            name = "id"
        )
    )
    """ The complete set of child roles that are associated
    with this role, this role should inherit all of the characteristics
    of the child roles (expected behaviour) """

    @classmethod
    def setup(cls):
        super(Role, cls).setup()

        # tries to find the owner role (default) in case it's not
        # found returns immediately nothing to be done
        owner = cls.find(name = "owner")
        if owner: return

        # retrieves the reference to the global logger that is going
        # to be used (should be initialized) and then prints the initial
        # information about the role to be generated
        logger = appier.get_logger()
        logger.info("Generating initial owner role for %s model ..." % cls.__name__)

        # creates the structure to be used as the owner role description
        # using the default value and then stores the role
        role = cls(
            name = "owner",
            description = "Super administrator role",
            tokens = ["*"]
        )
        role.save(validate = False)

    @classmethod
    def validate(cls):
        return super(Role, cls).validate() + [
            appier.not_null("name"),
            appier.not_empty("name"),
            appier.is_lower("name"),
            appier.string_gt("name", 3),
            appier.string_lt("name", 64),
            appier.not_duplicate("name", cls._name()),

            appier.not_null("tokens")
        ]

    @classmethod
    def list_names(cls):
        return ["name", "description"]

    def view_m(self, context = None):
        return self.view_

    @property
    def children_s(self):
        return [child for child in self.children if child and hasattr(child, "tokens_a")]

    @property
    def tokens_a(self):
        tokens = set(self.tokens)
        for child in self.children_s:
            tokens.update(child.tokens_a)
        return tokens

    @property
    def meta_a(self):
        meta = dict(self.meta)
        for child in self.children_s:
            self._join_m(child.meta_a, meta)
        return meta

    @property
    def secrets_a(self):
        secrets = dict(self.secrets)
        for child in self.children:
            self._join_m(child.secrets_a, secrets)
        return secrets

    def _join_m(self, origin, target):
        for key, value in appier.legacy.iteritems(origin):
            if key in target:
                previous = target[key]
                if isinstance(previous, dict):
                    value = self._join_m(value, previous)
                else:
                    if not isinstance(previous, list):
                        previous = [previous]
                    if not value in previous:
                        value = previous + [value]
                    else:
                        value = previous
            target[key] = value
        return target

    @appier.operation(
        name = "Duplicate",
        description = """Create a new account with exactly the
        same specification as the current one""",
        parameters = (("Suffix", "suffix", str, "-new"),),
        factory = True
    )
    def duplicate_s(self, suffix = "-new"):
        cls = self.__class__
        role = cls(
            description = self.description,
            meta = self.meta,
            name = self.name + suffix,
            tokens = self.tokens,
            view = self.view_
        )
        role.save()
        return role

    @appier.operation(
        name = "Set Parent",
        parameters = (("Name", "name", str),)
    )
    def set_parent_s(self, name):
        cls = self.__class__
        parent = cls.get(name = name)
        if self in parent.children: return
        parent.children.append(self)
        parent.save()

    @appier.operation(
        name = "Add Child(s)",
        parameters = (("Name", "name", str),)
    )
    def add_child_s(self, name):
        cls = self.__class__
        names = [name.strip() for name in name.strip().split(",")]
        for name in names:
            child = cls.get(name = name)
            if child in self.children: continue
            self.children.append(child)
        self.save()

    @appier.operation(
        name = "Remove Child(s)",
        parameters = (("Name", "name", str),)
    )
    def remove_child_s(self, name):
        cls = self.__class__
        names = [name.strip() for name in name.strip().split(",")]
        for name in names:
            child = cls.get(name = name)
            if not child in self.children: continue
            self.children.remove(child)
        self.save()

    @appier.operation(name = "Fix Children", level = 2)
    def fix_children_s(self):
        self.children = [child for child in self.children if child and hasattr(child, "tokens_a")]
        self.save()

    @appier.view(name = "Accounts")
    def accounts_v(self, *args, **kwargs):
        from appier_extras.parts.admin.models import account
        cls = account.Account
        kwargs["sort"] = kwargs.get("sort", [("id", -1)])
        kwargs.update(roles = {"$in" : [self.id]})
        return appier.lazy_dict(
            model = cls,
            kwargs = kwargs,
            entities = appier.lazy(lambda: cls.find(*args, **kwargs)),
            page = appier.lazy(lambda: cls.paginate(*args, **kwargs))
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import socket
import logging
import datetime
import threading

class LogglyHandler(logging.Handler):

    def __init__(
        self,
        level = logging.NOTSET,
        owner = None,
        api = None,
        buffer_size = 128,
        timeout = 30
    ):
        logging.Handler.__init__(self, level = level)
        self.owner = owner
        self.api = api
        self.buffer_size = buffer_size
        self.buffer = []
        self._schedule(timeout = timeout)

    def emit(self, record):
        # verifies if the API structure is defined and set and if
        # that's not the case returns immediately
        if not self.api: return

        # retrieves the current date time value as an utc value
        # and then formats it according to the provided format string
        now = datetime.datetime.utcnow()
        now_s = now.strftime("%Y-%m-%dT%H:%M:%S.%fZ")

        # creates the log record structure that is going to be sent
        # to the loggly infra-structure, this should represent a
        # proper structure ready to be debugged
        log = {
            "@timestamp" : now_s,
            "logger" : record.name,
            "message" : record.message,
            "level" : record.levelname,
            "path" : record.pathname,
            "lineno" : record.lineno,
            "host" : socket.gethostname(),
            "hostname" : socket.gethostname(),
            "tid" : threading.current_thread().ident,
            "pid" : os.getpid() if hasattr(os, "getpid") else -1,
        }

        # adds the new log item to the buffer so that it's properly
        # store for sending and then in case the should flush condition
        # is met the flush operation is performed, running the remote
        # calling operations that should block the infra-structure
        self.buffer.append(log)
        should_flush = len(self.buffer) >= self.buffer_size
        if should_flush: self.flush()

    def flush(self, force = False):
        logging.Handler.flush(self)

        # verifies if the API structure is defined and set and if
        # that's not the case returns immediately
        if not self.api: return

        # retrieves some references from the current instance that
        # are going to be used in the flush operation
        app = self.owner.owner
        buffer = self.buffer

        # verifies if the buffer is empty and if that's the case and
        # the force flag is not set, returns immediately
        if not buffer and not force: return

        # creates the lambda function that is going to be used for the
        # bulk flushing operation of the buffer, this is going to be
        # called on a delayed (async fashion) so that no blocking occurs
        # in the current logical flow
        call_log = lambda: self.api.log_bulk(buffer, tag = "default")

        # schedules the call log operation and then empties the buffer
        # so that it's no longer going to be used (flushed)
        app.delay(call_log)
        self.buffer = []

    def _schedule(self, timeout = 30):
        app = self.owner.owner

        def tick():
            self.flush()
            app.schedule(tick, timeout = timeout)

        app.schedule(tick, timeout = timeout)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import admin
from . import captcha
from . import contentful
from . import csfr
from . import diag
from . import export
from . import loggly
from . import logstash
from . import opbeat
from . import preflight
from . import prismic
from . import recaptcha
from . import sematext

from .admin import AdminPart
from .captcha import CaptchaPart
from .contentful import Contentful
from .csfr import CSFRPart, csfr_protect, csfr_ensure
from .diag import DiagPart
from .export import ExportPart
from .loggly import LogglyHandler, LogglyPart
from .logstash import LogstashHandler, LogstashPart
from .opbeat import OpbeatPart
from .preflight import PreflightPart
from .prismic import Prismic
from .recaptcha import ReCaptchaPart, recaptcha_protect, recaptcha_ensure
from .sematext import SematextHandler, SematextPart



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import hashlib

import appier

from appier_extras.parts.admin.models import base
from appier_extras.parts.admin.models.oauth import oauth_token

class OAuthClient(base.Base):
    """
    Represents an OAuth 2.0 API client and contains the
    secret information used to authenticate it. Should be
    able to generate access tokens.
    """

    name = appier.field(
        index = "hashed",
        default = True,
        immutable = True
    )
    """ The name of the client, this should be used as primary
    identification to the end-user """

    client_id = appier.field(
        index = "hashed",
        safe = True,
        immutable = True,
        description = "Client ID"
    )
    """ The client identifier issued to the client during the
    registration process (should be globally unique) """

    client_secret = appier.field(
        index = "hashed",
        safe = True,
        private = True,
        immutable = True
    )
    """ The client secret issued to the client during the
    registration process (should be globally unique) """

    redirect_uri = appier.field(
        index = "hashed",
        meta = "url",
        description = "Redirect URI"
    )
    """ The redirect URI where to redirect, the user agent
    after a successful token request operation """

    scope = appier.field(
        type = list
    )
    """ The sequence of scope values allowed for the current
    client, if not set all of the scope values are considered
    to be allowed """

    @classmethod
    def validate(cls):
        return super(OAuthClient, cls).validate() + [
            appier.not_null("name"),
            appier.not_empty("name"),
            appier.is_lower("name"),
            appier.string_gt("name", 3),
            appier.string_lt("name", 64),
            appier.not_duplicate("name", cls._name()),

            appier.not_null("redirect_uri"),
            appier.not_empty("redirect_uri"),
            appier.is_url("redirect_uri")
        ]

    @classmethod
    def list_names(cls):
        return ["name", "client_id", "redirect_uri"]

    @classmethod
    @appier.operation(
        name = "Create",
        parameters = (
            ("Name", "name", str),
            ("Redirect URI", "redirect_uri", str, "http://example.com/oauth")
        ),
        factory = True
    )
    def create_s(cls, name, redirect_uri):
        oauth_client = cls(
            name = name,
            redirect_uri = redirect_uri
        )
        oauth_client.save()
        return oauth_client

    @classmethod
    def _underscore(cls, plural = True):
        return "oauth_clients" if plural else "oauth_client"

    @classmethod
    def _readable(cls, plural = False):
        return "OAuth Clients" if plural else "OAuth Client"

    def pre_create(self):
        base.Base.pre_create(self)

        self.client_id = appier.gen_token(hash = hashlib.md5)
        self.client_secret = appier.gen_token()

    def assert_redirect_uri(self, redirect_uri):
        appier.verify(
            redirect_uri == self.redirect_uri,
            message = "The provided redirect URI is not valid",
            exception = appier.SecurityError,
            code = 403
        )

    def assert_scope(self, scope):
        if not self.scope: return
        invalid = [token for token in scope if not token in self.scope]
        appier.verify(
            not invalid,
            message = "The provided scope is not valid",
            exception = appier.SecurityError,
            code = 403
        )

    def get_tokens(self):
        return oauth_token.OAuthToken.find(client = self.id)

    @appier.operation(
        name = "Build Token",
        parameters = (
             ("Username", "username", str),
             ("Scope", "scope", list, ["me", "account"]),
             ("Redirect URI", "redirect_uri", str),
             ("Long Lived", "long", bool, False)
        ),
        factory = True,
        level = 2
    )
    def build_token_s(self, username = None, scope = [], redirect_uri = None, long = False):
        _oauth_token = oauth_token.OAuthToken()
        _oauth_token.username = username or self.session["username"]
        _oauth_token.scope = scope
        _oauth_token.redirect_uri = redirect_uri or self.redirect_uri
        _oauth_token.long = long
        _oauth_token.client = self
        _oauth_token.save()
        return _oauth_token

    @appier.operation(name = "Invalidate Tokens", level = 2)
    def invalidate_s(self):
        """
        Operation that invalidates the complete set of OAuth (access)
        tokens associated with the current OAuth client.

        This is a serious operation that should be used carefully to
        avoid unwanted behavior and loss of availability.
        """

        # retrieves the complete set of OAuth (access) tokens for the
        # current client and runs the delete operation (removing them)
        tokens = self.get_tokens()
        for token in tokens: token.delete()

    @appier.view(name = "Tokens")
    def tokens_v(self, *args, **kwargs):
        cls = oauth_token.OAuthToken
        kwargs["sort"] = kwargs.get("sort", [("id", -1)])
        kwargs.update(client = self.id)
        return appier.lazy_dict(
            model = cls,
            kwargs = kwargs,
            entities = appier.lazy(lambda: cls.find(*args, **kwargs)),
            page = appier.lazy(lambda: cls.paginate(*args, **kwargs))
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import unittest

import appier
import appier_extras

class OAuthTokenTest(unittest.TestCase):

    def setUp(self):
        self.app = appier.App(
            parts = (appier_extras.admin.AdminPart,),
            session_c = appier.MemorySession
        )

    def tearDown(self):
        self.app.unload()
        adapter = appier.get_adapter()
        adapter.drop_db()

    def test_basic(self):
        account = appier_extras.admin.Account()
        account.username = "username"
        account.email = "username@domain.com"
        account.password = "password"
        account.password_confirm = "password"
        account.save()

        oauth_client = appier_extras.admin.OAuthClient()
        oauth_client.name = "name"
        oauth_client.redirect_uri = "http://localhost/oauth"
        oauth_client.save()

        self.assertNotEqual(oauth_client.id, None)
        self.assertNotEqual(oauth_client.client_id, None)
        self.assertNotEqual(oauth_client.client_secret, None)
        self.assertEqual(oauth_client.name, "name")
        self.assertEqual(oauth_client.redirect_uri, "http://localhost/oauth")

        oauth_token = oauth_client.build_token_s(username = "username", scope = ["admin", "user"])

        self.assertNotEqual(oauth_token.id, None)
        self.assertNotEqual(oauth_token.access_token, None)
        self.assertNotEqual(oauth_token.authorization_code, None)
        self.assertNotEqual(oauth_token.authorization_code_date, None)
        self.assertEqual(oauth_token.client.id, oauth_client.id)
        self.assertEqual(oauth_token.username, "username")
        self.assertEqual(oauth_token.scope, ["admin", "user"])
        self.assertEqual(oauth_token.tokens, ["user"])

        tokens = oauth_client.get_tokens()

        self.assertEqual(len(tokens), 1)
        self.assertEqual(tokens[0].username, "username")

        oauth_client.invalidate_s()

        tokens = oauth_client.get_tokens()

        self.assertEqual(len(tokens), 0)

    def test_reuse(self):
        account = appier_extras.admin.Account()
        account.username = "username"
        account.email = "username@domain.com"
        account.password = "password"
        account.password_confirm = "password"
        account.save()

        oauth_client = appier_extras.admin.OAuthClient()
        oauth_client.name = "name"
        oauth_client.redirect_uri = "http://localhost/oauth"
        oauth_client.save()

        self.assertNotEqual(oauth_client.id, None)
        self.assertNotEqual(oauth_client.client_id, None)
        self.assertNotEqual(oauth_client.client_secret, None)
        self.assertEqual(oauth_client.name, "name")
        self.assertEqual(oauth_client.redirect_uri, "http://localhost/oauth")

        oauth_token = oauth_client.build_token_s(username = "username", scope = ["admin", "user"])

        self.assertNotEqual(oauth_token.id, None)
        self.assertNotEqual(oauth_token.access_token, None)
        self.assertNotEqual(oauth_token.authorization_code, None)
        self.assertNotEqual(oauth_token.authorization_code_date, None)
        self.assertEqual(oauth_token.client.id, oauth_client.id)
        self.assertEqual(oauth_token.username, "username")
        self.assertEqual(oauth_token.scope, ["admin", "user"])
        self.assertEqual(oauth_token.tokens, ["user"])

        result, tokens, oauth_token = appier_extras.admin.OAuthToken.reuse_s(
            "http://localhost/oauth",
            ["admin", "user"],
            oauth_client,
            account = account,
            owner = self.app
        )

        self.assertNotEqual(result, False)
        self.assertNotEqual(tokens, None)
        self.assertNotEqual(oauth_token, None)
        self.assertNotEqual(oauth_token.id, None)
        self.assertNotEqual(oauth_token.access_token, None)
        self.assertNotEqual(oauth_token.authorization_code, None)
        self.assertNotEqual(oauth_token.authorization_code_date, None)
        self.assertEqual(result, True)
        self.assertEqual(tokens, ["user"])
        self.assertEqual(oauth_token.client.id, oauth_client.id)
        self.assertEqual(oauth_token.username, "username")
        self.assertEqual(oauth_token.scope, ["admin", "user"])
        self.assertEqual(oauth_token.tokens, ["user"])

    def test_change_username(self):
        account = appier_extras.admin.Account()
        account.username = "username"
        account.email = "username@domain.com"
        account.password = "password"
        account.password_confirm = "password"
        account.save()

        oauth_client = appier_extras.admin.OAuthClient()
        oauth_client.name = "name"
        oauth_client.redirect_uri = "http://localhost/oauth"
        oauth_client.save()

        self.assertNotEqual(oauth_client.id, None)
        self.assertNotEqual(oauth_client.client_id, None)
        self.assertNotEqual(oauth_client.client_secret, None)
        self.assertEqual(oauth_client.name, "name")
        self.assertEqual(oauth_client.redirect_uri, "http://localhost/oauth")

        oauth_token = oauth_client.build_token_s(username = "username", scope = ["admin", "user"])

        self.assertNotEqual(oauth_token.id, None)
        self.assertNotEqual(oauth_token.access_token, None)
        self.assertNotEqual(oauth_token.authorization_code, None)
        self.assertNotEqual(oauth_token.authorization_code_date, None)
        self.assertEqual(oauth_token.client.id, oauth_client.id)
        self.assertEqual(oauth_token.username, "username")
        self.assertEqual(oauth_token.scope, ["admin", "user"])
        self.assertEqual(oauth_token.tokens, ["user"])

        account = account.reload()
        account.change_username_s("username_changed")

        oauth_token_p = oauth_token
        oauth_token = oauth_token.reload(rules = False)

        self.assertNotEqual(oauth_token.id, None)
        self.assertNotEqual(oauth_token.access_token, None)
        self.assertNotEqual(oauth_token.authorization_code, None)
        self.assertNotEqual(oauth_token.authorization_code_date, None)
        self.assertEqual(oauth_token.client.id, oauth_client.id)
        self.assertEqual(oauth_token.username, "username_changed")
        self.assertEqual(oauth_token.scope, ["admin", "user"])
        self.assertEqual(oauth_token.tokens, ["user"])

        self.assertNotEqual(id(oauth_token), id(oauth_token_p))
        self.assertEqual(oauth_token.id, oauth_token_p.id)
        self.assertEqual(oauth_token.access_token, oauth_token_p.access_token)
        self.assertEqual(oauth_token.authorization_code, oauth_token_p.authorization_code)
        self.assertEqual(oauth_token.authorization_code_date, oauth_token_p.authorization_code_date)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2021 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2021 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import json
import socket
import datetime

import appier

from appier_extras import utils
from appier_extras.parts.admin.models import base

class Event(base.Base):

    name = appier.field(
        index = "all",
        default = True
    )

    handler = appier.field(
        index = "all"
    )

    arguments = appier.field(
        type = dict,
        meta = "longmap"
    )

    @classmethod
    def validate(cls):
        return super(Event, cls).validate() + [
            appier.not_null("name"),
            appier.not_empty("name"),

            appier.not_null("handler"),
            appier.not_empty("handler")
        ]

    @classmethod
    def list_names(cls):
        return ["name", "handler", "enabled"]

    @classmethod
    def transform(cls, arguments, prefix):
        """
        Runs the transformation process on the provided arguments
        taking into account the filtering prefix, so that if the
        argument name is prefixed by the provided prefix such prefix
        is removed on a new re-set arguments.

        :type arguments: Dictionary
        :param arguments: The dictionary of arguments for which the
        prefix transformation is going to take place.
        :type prefix: String
        :param prefix: The prefix that is going to be used in the
        prefix removal transformation process.
        :rtype: Dictionary
        :return: The new dictionary of arguments with the processed
        values without the associated prefix set.
        """

        arguments = dict(arguments)
        prefix_l = len(prefix)
        for key, value in appier.legacy.items(arguments):
            if not key.startswith(prefix): continue
            key_prefix = key[prefix_l:]
            arguments[key_prefix] = value
        return arguments

    @classmethod
    def format(cls, arguments, all = False):
        formatter = utils.SafeFormatter()
        arguments = dict(arguments)
        for key, value in appier.legacy.items(arguments):
            if not appier.legacy.is_string(value, all = all): continue
            try: value = formatter.format(value, **arguments)
            except Exception: value = value
            arguments[key] = value
        return arguments

    @classmethod
    def notify_g(cls, name, handlers = None, permutations = True, arguments = {}):
        logger = appier.get_logger()
        logger.debug("Triggering '%s' event ..." % name)
        if permutations: names = cls._get_permutations(name)
        else: names = (name,)
        kwargs = dict()
        kwargs["name"] = {"$in": names}
        if handlers: kwargs["handler"] = {"$in" : handlers}
        events = cls.find_e(**kwargs)
        for event in events:
            event.name = name
            event.notify(arguments = arguments)

    @appier.operation(name = "Notify")
    def notify(self, arguments = {}, delay = True):
        cls = self.__class__
        delay_s = "a delayed" if delay else "an immediate"
        logger = appier.get_logger()
        logger.debug(
            "Notifying handler '%s' for '%s' in %s fashion ..." %\
            (self.handler, self.name, delay_s)
        )
        method = getattr(self, "notify_" + self.handler)
        arguments_m = dict(self.arguments)
        arguments_m.update(arguments)
        arguments_m.update(event = self.name, handler = self.handler)
        arguments_m = cls.transform(arguments_m, self.handler + "_")
        arguments_m = cls.format(arguments_m)
        kwargs = dict(arguments = arguments_m)
        if delay: self.owner.delay(method, kwargs = kwargs)
        else: method(**kwargs)

    @classmethod
    @appier.operation(
        name = "Import CSV",
        parameters = (
            ("CSV File", "file", "file"),
            ("Empty source", "empty", bool, False)
        )
    )
    def import_csv_s(cls, file, empty):

        def callback(line):
            name, handler, arguments = line
            name = name or None
            handler = handler or None
            arguments = json.loads(arguments) if arguments else None
            event = cls(
                name = name,
                handler = handler,
                arguments = arguments
            )
            event.save()

        if empty: cls.delete_c()
        cls._csv_import(file, callback)

    @classmethod
    @appier.link(name = "Export CSV")
    def list_csv_url(cls, absolute = False):
        return appier.get_app().url_for(
            "admin.list_events_csv",
            absolute = absolute
        )

    @classmethod
    def _retry(cls, callable, count = 3):
        result = None
        while True:
            try:
                result = callable()
            except Exception:
                count -= 1
                if count == 0: raise
                else: continue
            break
        return result

    def notify_http(self, arguments = {}):
        cls = self.__class__
        url = arguments.get("url", None)
        retries = arguments.get("retries", 3)
        logger = appier.get_logger()
        logger.debug("Running HTTP notification for '%s' ..." % url)
        return cls._retry(lambda: appier.post(url, data_j = arguments), count = retries)

    def notify_mailme(self, arguments = {}):
        cls = self.__class__
        appier.ensure_pip("mailme", package = "mailme_api")
        import mailme
        retries = arguments.get("retries", 3)
        logger = appier.get_logger()
        logger.debug("Running Mailme notification ...")
        api = mailme.API()
        return cls._retry(lambda: api.send(arguments), count = retries)

    def notify_nexmo(self, arguments = {}):
        cls = self.__class__
        appier.ensure_pip("nexmo", package = "nexmo_api")
        import nexmo
        sender = arguments["sender"]
        receiver = arguments["receiver"]
        text = arguments["text"]
        retries = arguments.get("retries", 3)
        logger = appier.get_logger()
        logger.debug("Running Nexmo notification for '%s' ..." % receiver)
        api = nexmo.API()
        return cls._retry(lambda: api.send_sms(sender, receiver, text), count = retries)

    @classmethod
    def notify_kafka(cls, arguments = {}):
        producer = cls._get_kafka()

        name = arguments["event"]
        params = arguments["params"]
        payload = params.get("payload", dict())

        if "topic" in arguments: topic = arguments["topic"]
        else: topic = name.split(".")[0]

        data = json.dumps(
            dict(
                name = name,
                origin = appier.get_name(),
                hostname = socket.gethostname(),
                datatype = "json",
                timestamp = datetime.datetime.now(),
                payload = payload
            )
        )
        data_b = appier.legacy.bytes(data, encoding = "utf-8", force = True)

        producer.send(topic, data_b)
        producer.flush()

    @appier.operation(name = "Duplicate", factory = True)
    def duplicate_s(self):
        cls = self.__class__
        event = cls(
            name = self.name,
            handler = self.handler,
            arguments = self.arguments
        )
        event.save()
        return event

    @classmethod
    def _get_kafka(cls):
        if hasattr(cls, "_kafka"): return cls._kafka

        appier.ensure_pip("kafka", package = "kafka-python")
        import kafka

        kafka_server = appier.conf("KAFKA_SERVER", "localhost:9092")
        client_id = appier.conf("KAFKA_PRODUCER_CLIENT_ID", None)
        compression_type = appier.conf("KAFKA_COMPRESSION_TYPE", None)
        retries = appier.conf("KAFKA_RETRIES", 0, cast = int)
        batch_size = appier.conf("KAFKA_BATCH_SIZE", 16384, cast = int)
        linger_ms = appier.conf("KAFKA_LINGER", 0, cast = int)
        connections_max_idle_ms = appier.conf("KAFKA_CONNECTIONS_MAX_IDLE", 540000, cast = int)
        max_request_size = appier.conf("KAFKA_MAX_REQUEST_SIZE", 1048576, cast = int)
        retry_backoff_ms = appier.conf("KAFKA_RETRY_BACKOFF", 100, cast = int)
        request_timeout_ms = appier.conf("KAFKA_REQUEST_TIMEOUT", 30000, cast = int)

        cls._kafka = kafka.KafkaProducer(
            bootstrap_servers = kafka_server,
            client_id = client_id,
            compression_type = compression_type,
            retries = retries,
            batch_size = batch_size,
            linger_ms = linger_ms,
            connections_max_idle_ms = connections_max_idle_ms,
            max_request_size = max_request_size,
            retry_backoff_ms = retry_backoff_ms,
            request_timeout_ms = request_timeout_ms
        )

        return cls._kafka

    @classmethod
    def _get_permutations(cls, name, wildcard = "*"):
        names = [wildcard]
        names_a = name.split(".")[:-1]
        for name_i in range(len(names_a)):
            name_join = ".".join([n for n in names_a[0:name_i + 1]])
            names.append("%s.%s" % (name_join, wildcard))
        names.append(name)
        return names
